# This file is part of the project mb-netmgmt
#
# (C) 2022 Deutsche Telekom AG
#
# Deutsche Telekom AG and all other contributors / copyright
# owners license this file to you under the terms of the GPL-2.0:
#
# mb-netmgmt is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# mb-netmgmt is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mb-netmgmt. If not, see <https://www.gnu.org/licenses/

import binascii
from base64 import b64decode, b64encode
from socket import SOCK_DGRAM, socket
from socketserver import DatagramRequestHandler, ThreadingUDPServer, UDPServer

from scapy.asn1.asn1 import ASN1_Class_UNIVERSAL
from scapy.layers.snmp import SNMP, SNMPbulk, SNMPresponse, SNMPvarbind

from mb_netmgmt.__main__ import Protocol

Server = ThreadingUDPServer


class Handler(DatagramRequestHandler, Protocol):
    def handle(self):
        self.callback_url = self.server.callback_url
        request, request_id = self.read_request()
        self.handle_request(request, request_id)

    def handle_request(self, request, request_id):
        mb_response = self.post_request(request)
        if "response" not in mb_response:
            self.open_upstream(mb_response["proxy"]["to"])
            self.send_upstream(request, request_id)
        response = self.get_response(mb_response)
        self.respond(response, request_id)

    def respond(self, response, request_id):
        response_varbindlist = self.translate_json_to_network_response(response)
        snmp_response = SNMP(
            PDU=SNMPresponse(
                id=request_id,
                varbindlist=response_varbindlist,
            )
        )
        self.wfile.write(bytes(snmp_response))

    def open_upstream(self, to):
        self.upstream_socket = socket(type=SOCK_DGRAM)
        self.upstream_socket.connect((to, 161))

    def read_proxy_response(self):
        bytes_response = self.upstream_socket.recv(UDPServer.max_packet_size)
        snmp_response = SNMP(bytes_response)
        result = dict()
        for varbind in snmp_response.PDU.varbindlist:
            result[varbind.oid.val] = {
                "val": decode(varbind.value.val),
                "tag": str(varbind.value.tag),
            }

        return result

    def send_upstream(self, request, request_id):
        oids = request["oids"]
        pdu_type = type(self.snmp_request.PDU)
        kwargs = dict()
        if pdu_type == SNMPbulk:
            kwargs["max_repetitions"] = 10
        request = SNMP(
            version="v2c",
            community=self.snmp_request.community,
            PDU=pdu_type(
                id=self.snmp_request.PDU.id,
                varbindlist=[SNMPvarbind(oid=oid) for oid in oids],
                **kwargs,
            ),
        )
        self.upstream_socket.send(bytes(request))

    def read_request(self):
        self.snmp_request = SNMP(self.rfile.read())
        pdu_id = self.snmp_request.PDU.id.val
        json_request = {
            "oids": [varbind.oid.val for varbind in self.snmp_request.PDU.varbindlist]
        }
        return json_request, pdu_id

    def translate_request_to_json(self, varbind):
        return {"oid": varbind.oid.val}

    def translate_json_to_network_response(self, json):
        result = list()
        for oid, response in json.items():
            if oid.startswith("_"):
                continue
            value = response["val"]
            try:
                value = b64decode(value, validate=True)
            except (binascii.Error, TypeError):
                pass
            asn1_class = ASN1_Class_UNIVERSAL.__dict__[response["tag"]]
            result += SNMPvarbind(oid=oid, value=asn1_class.asn1_object(value))
        return result


def decode(value):
    try:
        return value.decode()
    except AttributeError:
        return value
    except UnicodeDecodeError:
        return b64encode(value).decode()



# This file is part of the project mb-netmgmt
#
# (C) 2022 Deutsche Telekom AG
#
# Deutsche Telekom AG and all other contributors / copyright
# owners license this file to you under the terms of the GPL-2.0:
#
# mb-netmgmt is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# mb-netmgmt is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mb-netmgmt. If not, see <https://www.gnu.org/licenses/

from socketserver import BaseRequestHandler
from socketserver import TCPServer as Server

import paramiko

from mb_netmgmt.__main__ import Protocol

stopped = False


class ParamikoServer(paramiko.ServerInterface):
    def get_allowed_auths(*args):
        return "password,publickey"

    def check_auth_password(*args):
        return paramiko.AUTH_SUCCESSFUL

    def check_auth_publickey(*args):
        return paramiko.AUTH_SUCCESSFUL

    def check_channel_request(*args):
        return paramiko.OPEN_SUCCEEDED

    def check_channel_pty_request(*args):
        return True

    def check_channel_shell_request(*args):
        return True

    def check_channel_subsystem_request(*args):
        return True


class Handler(BaseRequestHandler, Protocol):
    message_terminators = [b"\n"]
    default_port = 22

    def handle(self):
        self.callback_url = self.server.callback_url
        transport = start_server(self.request)
        self.channel = transport.accept()
        self.open_upstream()
        self.handle_prompt()
        while not stopped:
            request, request_id = self.read_request()
            self.handle_request(request, request_id)

    def send_upstream(self, request, request_id):
        self.upstream_channel.sendall(request["command"])

    def read_request(self):
        request = self.read_message(self.channel)
        return {"command": request.decode()}, None

    def respond(self, response, request_id):
        response = response["response"]
        self.channel.sendall(response)

    def open_upstream(self):
        to = self.get_to()
        if not to:
            return
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy)
        client.connect(
            to.hostname,
            to.port or self.default_port,
            to.username,
            to.password,
            key_filename=self.keyfile.name,
            transport_factory=paramiko.Transport,
        )
        self.upstream_channel = client.invoke_shell()

    def handle_prompt(self):
        self.handle_request({"command": ""}, "")

    def read_proxy_response(self):
        return {"response": self.read_message(self.upstream_channel).decode()}

    def read_message(self, channel):
        message = b""
        end_of_message = False
        while not end_of_message and not stopped:
            message += channel.recv(1024)
            for terminator in self.message_terminators:
                if terminator in message:
                    end_of_message = True
        return message


def start_server(request):
    t = paramiko.Transport(request)
    t.add_server_key(paramiko.DSSKey.generate())
    t.add_server_key(paramiko.ECDSAKey.generate())
    t.add_server_key(paramiko.RSAKey.generate(4096))
    t.start_server(server=ParamikoServer())
    return t



# This file is part of the project mb-netmgmt
#
# (C) 2022 Deutsche Telekom AG
#
# Deutsche Telekom AG and all other contributors / copyright
# owners license this file to you under the terms of the GPL-2.0:
#
# mb-netmgmt is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# mb-netmgmt is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mb-netmgmt. If not, see <https://www.gnu.org/licenses/

import logging
from socketserver import BaseRequestHandler
from socketserver import TCPServer as Server

from lxml import etree
from ncclient.devices.default import DefaultDeviceHandler
from ncclient.manager import connect
from ncclient.transport.parser import DefaultXMLParser
from ncclient.transport.session import (
    BASE_NS_1_0,
    HelloHandler,
    NetconfBase,
    SessionListener,
    qualify,
    sub_ele,
    to_ele,
)
from ncclient.transport.ssh import MSG_DELIM, PORT_NETCONF_DEFAULT, SSHSession

from mb_netmgmt.__main__ import Protocol
from mb_netmgmt.ssh import start_server

stopped = False


class Handler(BaseRequestHandler, Protocol):
    def setup(self):
        session = SSHSession(DefaultDeviceHandler())
        session.add_listener(Listener(super().handle_request))
        self.parser = DefaultXMLParser(session)
        self.session = session

    def handle(self):
        self.callback_url = self.server.callback_url
        transport = start_server(self.request)
        self.channel = transport.accept()
        self.open_upstream()
        self.session._connected = True
        self.handle_prompt()
        self.session._transport = transport
        self.session._channel = self.channel
        self.session.run()

    def open_upstream(self):
        to = self.get_to()
        if not to:
            return
        self.manager = connect(
            host=to.hostname,
            port=to.port or PORT_NETCONF_DEFAULT,
            username=to.username,
            password=to.password,
            key_filename=self.keyfile.name,
            hostkey_verify=False,
            timeout=60,
        )

    def handle_prompt(self):
        hello = to_ele(
            HelloHandler.build(DefaultDeviceHandler._BASE_CAPABILITIES, None)
        )

        # A server sending the <hello> element MUST include a <session-id>
        # element containing the session ID for this NETCONF session.
        # https://datatracker.ietf.org/doc/html/rfc6241#section-8.1
        session_id = sub_ele(hello, "session-id")
        session_id.text = "1"

        self.channel.sendall(to_xml(hello) + MSG_DELIM.decode())

        def init_cb(id, capabilities):
            if "urn:ietf:params:netconf:base:1.1" in capabilities:
                self.session._base = NetconfBase.BASE_11

        self.session.add_listener(HelloHandler(init_cb, lambda ex: None))

    def read_proxy_response(self):
        return {"rpc-reply": remove_message_id(self.rpc_reply._root)}

    def send_upstream(self, request, request_id):
        self.rpc_reply = self.manager.rpc(to_ele(request["rpc"]))

    def respond(self, response, request_id):
        reply = response.get("rpc-reply", f'<rpc-reply xmlns="{BASE_NS_1_0}"/>')
        message = add_message_id(reply, request_id)
        self.session.send(message)


class Listener(SessionListener):
    def __init__(self, handle_request):
        self.handle_request = handle_request

    def callback(self, root, raw):
        tag, attrs = root
        if (tag == qualify("hello")) or (tag == "hello"):
            return
        ele = etree.fromstring(raw.encode())
        request = {"rpc": to_xml(ele[0])}
        self.handle_request(request, attrs["message-id"])

    def errback(self, ex):
        logging.exception(ex)


def add_message_id(rpc_reply, message_id):
    ele = etree.fromstring(rpc_reply)
    ele.set("message-id", message_id)
    return to_xml(ele)


def remove_message_id(root):
    del root.attrib["message-id"]
    return to_xml(root)


def to_xml(ele):
    return etree.tostring(ele, pretty_print=True).decode().strip()



import os
from threading import Thread
from urllib.parse import urlparse

import ncclient.manager
import paramiko
import pytest
from mb_netmgmt import mb, netconf, ssh
from mb_netmgmt.__main__ import create_server
from ncclient.transport.session import BASE_NS_1_0, to_ele
from ncclient.transport.ssh import MSG_DELIM

port = 8080
prompt = b"prompt"
mock_response = f"""<rpc-reply xmlns="{BASE_NS_1_0}">
  <blubb/>
</rpc-reply>"""


@pytest.mark.parametrize("protocol", ["http", "snmp", "telnet", "netconf"])
def test_create_imposter(protocol):
    with mb(imposter(protocol, None)):
        pass


def test_ssh():
    with mb(imposter("ssh", [prompt_stub()]), "debug"):
        client = connect_ssh()
        chan = client.invoke_shell()
        out = chan.recv(1024)
        assert out == prompt


def test_ssh_proxy():
    with mb(
        imposter(
            "ssh",
            [
                prompt_stub(),
                {
                    "responses": [
                        {
                            "proxy": {
                                "to": f"ssh://{os.environ['NETCONF_USERNAME']}:{os.environ['NETCONF_PASSWORD']}@localhost"
                            }
                        },
                    ]
                },
            ],
        )
    ):
        client = connect_ssh()
        chan = client.invoke_shell()
        out = chan.recv(1024)
        assert out == prompt


def prompt_stub():
    return {"responses": [{"is": {"response": prompt}}]}


def test_create_ssh_server():
    server = create_server(ssh, port, None)
    Thread(target=server.serve_forever).start()
    connect_ssh().close()
    server.shutdown()


def connect_ssh():
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy)
    client.connect("localhost", port)
    return client


def test_create_netconf_server():
    port = 8830
    original_open_upstream = netconf.Handler.open_upstream
    original_post_request = netconf.Handler.post_request
    netconf.Handler.open_upstream = lambda handler: None
    netconf.Handler.post_request = mock_post_request
    server = create_server(netconf, port, None)
    Thread(target=server.serve_forever).start()
    with ncclient.manager.connect(
        host="localhost", port=port, password="", hostkey_verify=False
    ) as m:
        assert m.session_id == "1"
        assert "urn:ietf:params:netconf:base:1.0" in m.server_capabilities
        m.get_config("running")
    netconf.stopped = True
    server.shutdown()
    netconf.stopped = False
    netconf.Handler.open_upstream = original_open_upstream
    netconf.Handler.post_request = original_post_request


def mock_post_request(handler, request):
    return {"response": {"rpc-reply": mock_response}}


def test_netconf_upstream():
    from mb_netmgmt.netconf import Handler

    global port
    port = 8831
    Handler.handle = lambda handler: None
    Handler.get_to = lambda handler: urlparse(f"netconf://localhost:{port}")
    handler = Handler(None, None, None)
    handler.save_key({})
    with mb(
        imposter(
            "netconf",
            [
                {
                    "predicates": [{"endsWith": {"rpc": ">running</get-config>"}}],
                    "responses": [
                        {"is": {"rpc-reply": mock_response}},
                    ],
                }
            ],
        ),
        "debug",
    ):
        handler.open_upstream()
        handler.send_upstream(
            {"rpc": "<get-config>running</get-config>"},
            42,
        )
        proxy_response = handler.read_proxy_response()
        assert proxy_response["rpc-reply"] == mock_response


def imposter(protocol, stubs=None):
    return [{"protocol": protocol, "port": port, "stubs": stubs}]


def create_proxy_response(message_id):
    return {
        "response": f'<rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="{message_id}"/>'
        + MSG_DELIM.decode()
    }


def test_remove_message_id():
    result = netconf.remove_message_id(
        to_ele(f'<rpc-reply xmlns="{BASE_NS_1_0}" message-id="42"><blubb/></rpc-reply>')
    )
    assert result == mock_response


def test_netconf_default_response():
    with mb(imposter("netconf")):
        with ncclient.manager.connect(
            host="localhost", port=port, password="", hostkey_verify=False
        ) as m:
            m.get_config("running")


def test_netconf_private_key():
    with mb(
        imposter(
            "netconf",
            [
                {
                    "responses": [
                        {
                            "proxy": {
                                "to": f"netconf://{os.environ['NETCONF_USERNAME']}@{os.environ['NETCONF_HOSTNAME']}",
                                "key": os.environ["NETCONF_KEY"],
                            }
                        },
                    ]
                },
            ],
        )
    ):
        ncclient.manager.connect(
            host="localhost", port=port, password="", hostkey_verify=False
        )
